#!/usr/bin/env python3
# Copyright    2023  The University of Electro-Communications  (Author: Teo Wen Shen)  # noqa
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Corpus owner: https://clrd.ninjal.ac.jp/csj/en/index.html
Corpus description:
- http://www.lrec-conf.org/proceedings/lrec2000/pdf/262.pdf
- https://isca-speech.org/archive_open/archive_papers/sspr2003/sspr_mmo2.pdf

This script accesses the CSJ/SDB/{core,noncore} directories, and generates
transcripts where each word is formatted as '{surface}+{morph}+{pron}'.

NOTE: Besides the eval1, eval2, eval3, and excluded datasets introduced in
kaldi, this script also creates a separate validation dataset. Recording
sessions chosen for validation are listed in this script.

This script does the following in sequence:-

**MOVE**
0. This stage is skipped if `--transcript-dir` is not provided.
1. Copies each .sdb files from /SDB into its own directory in the designated
  `transcript_dir`, i.e. {transcript_dir}/{spk_id}/{spk_id}.sdb
2. Verifies that the corresponding wav file exists in the /WAV directory, and
   outputs that absolute path into {spk_id}-wav.list
3. Moves the predefined datasets for eval1, eval2, eval3, and excluded, into
   its own dataset directory
4. Touches a .done_mv in `transcript_dir`.
NOTE: If a .done_mv exists already in `transcript_dir`, then this stage is skipped.

**PREPARE MANIFESTS**
1. Parses all .sdb files it can find within `transcript_dir` in disfluent mode.
2. Generates supervisions and recordings manifests for each dataset part.

Differences to kaldi include:-
1. The evaluation datasets do not follow `trans_dir`/eval/eval{i}, but are
   instead saved in the same level as core, noncore, and excluded.
2. A validation dataset is explicitly specified.
3. Utterances with "×" are not removed. You will need to remove them in a
   later stage.
4. Segments are not concatenated, unless an F-tag, D-tag, L-tag, or A-tag
   spans between two segments.
5. Multi-segment M-tags, R-tags, and O-tags are removed, while M-tags,
   R-tags, and O-tags within a segment are retained.

Example structure of SupervisionSegment.custom:-
{
  "raw": (
      "(F_えー)+感動詞+(F_エー) (M_(F_うーん)+感動詞+(M_(F_(W_ウー;ウーン)) "
      "それ+代名詞+ソレ だっ+助動詞/促音便+ダッ たら)+助動詞+タラ) と+助詞/格助詞+ト "
      "いう+動詞/ワア行五段+ユー の+助詞/準体助詞+ノ は+助詞/係助詞+ワ "
      "(F_えー)+感動詞+(F_エー)"
      ),
  "disfluent": "えーうーんそれだったらというのはえー",
  "disfluent_tag": "F,F,M/F,M/F,M/F,M,M,M,M,M,M,,,,,,F,F"
}
NOTE:
1. XX_tag is guaranteed to be the same length as XX. It labels the tag to which each
   character belongs. It is useful for evaluation.
2. The SupervisionSegment.text field is populated with 'disfluent', i.e.
   SupervisionSegment.text == SupervisionSegment.custom['dislfuent'], so that this
   supervision is compatible with other recipes.

The transcript directory, if generated, has this structure:-
{transcript_dir}
 - excluded
   - ...
 - core
   - ...
 - eval1
   - ...
 - eval2
   - ...
 - eval3
   - ...
 - noncore
   - ...
   - A01F0576
     - A01F0576.sdb
     - A01F0576-trans.txt
     - A01F0576-wav.list
   - ...
   - D03M0038
     - D03M0038.sdb
     - D03M0038-L-trans.txt
     - D03M0038-L-wav.list
     - D03M0038-R-trans.txt
     - D03M0038-R-wav.list
 - valid
     - ...

"""

import copy
import logging
import re
from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Callable, Dict, List, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike

# ---------------------------------- #
#    Prepare transcripts from SDB    #


_FULL_DATA_PARTS = ["eval1", "eval2", "eval3", "excluded", "valid", "core", "noncore"]

# Exclude speaker ID
_A01M0056 = [
    "S05M0613",
    "R00M0187",
    "D01M0019",
    "D04M0056",
    "D02M0028",
    "D03M0017",
]

_VALID = [
    "A01M0264",
    "A01M0377",
    "A01M0776",
    "A01M0891",
    "A03F0109",
    "A04M0899",
    "A05M0420",
    "A07M0318",
    "A07M0912",
    "A11M0795",
    "A12M0983",
    "D03F0058",
    "R00M0415",
    "R01F0101",
    "R01F0125",
    "R02M0073",
    "R03F0108",
    "R03F0157",
    "S00F0014",
    "S00M0793",
    "S01F0507",
    "S02F0122",
    "S02F0362",
    "S02M1351",
    "S02M1372",
    "S03F1199",
    "S04F1020",
    "S05F0443",
    "S07F0853",
    "S07F1333",
    "S07M0827",
    "S08F0717",
    "S08F1340",
    "S09M0619",
    "S10M1090",
    "S10M1275",
    "S11F0578",
    "S11M0864",
    "S11M1174",
]

# Evaluation set ID
_EVAL = [
    # eval1
    [
        "A01M0110",
        "A01M0137",
        "A01M0097",
        "A04M0123",
        "A04M0121",
        "A04M0051",
        "A03M0156",
        "A03M0112",
        "A03M0106",
        "A05M0011",
    ],
    # eval2
    [
        "A01M0056",
        "A03F0072",
        "A02M0012",
        "A03M0016",
        "A06M0064",
        "A06F0135",
        "A01F0034",
        "A01F0063",
        "A01F0001",
        "A01M0141",
    ],
    # eval3
    [
        "S00M0112",
        "S00F0066",
        "S00M0213",
        "S00F0019",
        "S00M0079",
        "S01F0105",
        "S00F0152",
        "S00M0070",
        "S00M0008",
        "S00F0148",
    ],
]


def _create_trans_dir(corpus_dir: Path, trans_dir: Path):

    if (trans_dir / ".done_mv").exists():
        logging.info(
            f"{trans_dir} already created. "
            f"Delete {trans_dir / '.done_mv'} to create again."
        )
        return

    for ori_files in (corpus_dir / "MORPH/SDB").glob("*/*.sdb"):
        vol = ori_files.parts[-2]
        spk_id = ori_files.name[:-4]
        new_dir = trans_dir / vol / spk_id
        new_dir.mkdir(parents=True, exist_ok=True)
        wav_dir = corpus_dir / "WAV" / vol

        if spk_id[0] == "D":
            l_wav = wav_dir / f"{spk_id}-L.wav"
            r_wav = wav_dir / f"{spk_id}-R.wav"
            assert l_wav.is_file(), f"{spk_id}-L.wav cannot be found"
            assert r_wav.is_file(), f"{spk_id}-R.wav cannot be found"
            (new_dir / f"{spk_id}-L-wav.list").write_text(
                l_wav.as_posix(), encoding="utf8"
            )
            (new_dir / f"{spk_id}-R-wav.list").write_text(
                r_wav.as_posix(), encoding="utf8"
            )
            L_sdb = []
            R_sdb = []
            for line in ori_files.read_text(encoding="shift_jis").split("\n"):
                if not line:
                    L_sdb.append(line)
                    R_sdb.append(line)
                elif "L:" in line.split("\t")[3]:
                    L_sdb.append(line)
                else:
                    assert "R:" in line, line
                    R_sdb.append(line)
            (new_dir / f"{spk_id}-R.sdb").write_text(
                "\n".join(R_sdb), encoding="shift_jis"
            )
            (new_dir / f"{spk_id}-L.sdb").write_text(
                "\n".join(L_sdb), encoding="shift_jis"
            )
        else:
            (new_dir / f"{spk_id}.sdb").write_bytes(ori_files.read_bytes())
            wav = wav_dir / f"{spk_id}.wav"
            assert wav.is_file(), f"{spk_id}.wav cannot be found"
            (new_dir / f"{spk_id}-wav.list").write_text(wav.as_posix(), encoding="utf8")

    for ori_files in _A01M0056:
        ori_files = list(trans_dir.glob(f"*/{ori_files}/{ori_files}*"))

        for ori_file in ori_files:
            *same_part, vol, spk_id, filename = ori_file.as_posix().split("/")
            new_dir = Path("/".join(same_part + ["excluded", spk_id]))
            new_dir.mkdir(parents=True, exist_ok=True)
            ori_file.rename(new_dir / filename)
        ori_files[0].parent.rmdir()

    for i, eval_list in enumerate(_EVAL, start=1):
        for ori_files in eval_list:
            ori_files = list(trans_dir.glob(f"*/{ori_files}/{ori_files}*"))

            for ori_file in ori_files:
                *same_part, vol, spk_id, filename = ori_file.as_posix().split("/")
                new_dir = Path("/".join(same_part + [f"eval{i}", spk_id]))
                new_dir.mkdir(parents=True, exist_ok=True)
                ori_file.rename(new_dir / filename)
            ori_files[0].parent.rmdir()

    for ori_files in _VALID:
        ori_files = list(trans_dir.glob(f"*/{ori_files}/{ori_files}*"))

        for ori_file in ori_files:
            *same_part, vol, spk_id, filename = ori_file.as_posix().split("/")
            new_dir = Path("/".join(same_part + ["valid", spk_id]))
            new_dir.mkdir(parents=True, exist_ok=True)
            ori_file.rename(new_dir / filename)
        ori_files[0].parent.rmdir()

    (trans_dir / ".done_mv").touch()
    logging.info("Transcripts have been moved.")


# ----------------------------

INTERNAL_SEP = " "
_FIELDS = {
    "time": 3,
    "surface": 5,
    "notag": 9,
    "pos1": 11,
    "cForm": 12,
    "cType1": 13,
    "pos2": 14,
    "cType2": 15,
    "other": 16,
    "pron": 10,
    "spkid": 2,
}

_MORPH = ["pos1", "cForm", "cType2", "pos2"]

_REPLACEMENTS = [
    "<FV>",
    "<VN>",
    "<H>",
    "<Q>",
    "<笑>",
    "<咳>",
    "<息>",
    "<泣>",
    "<フロア発話>",
    "<フロア笑>",
    "<拍手>",
    "<デモ>",
    "<ベル>",
    "<朗読間違い>",
    "<雑音>",
]

DECISIONS = {
    "F": 0,
    "D": 0,
    "D2": 0,
    "?": 0,
    "?,": 0,
    "M": 0,
    "O": 0,
    "R": 0,
    "X": 0,
    "A": 1,
    "A_num": 0,
    "K": 1,
    "W": 1,
    "B": 0,
    "笑": 0,
    "泣": 0,
    "咳": 0,
    "L": 0,
}


class _CSJSDBWord:
    time = ""
    surface = ""
    notag = ""
    pos1 = ""
    cForm = ""
    cType1 = ""
    pos2 = ""
    cType2 = ""
    other = ""
    pron = ""
    spkid = ""
    sgid = 0
    start = -1.0
    end = -1.0
    morph = ""

    @staticmethod
    def from_line(line=""):
        word = _CSJSDBWord()
        line = line.strip().split("\t")

        for f, i in _FIELDS.items():
            try:
                attr = line[i]
            except IndexError:
                attr = ""
            setattr(word, f, attr)

        for _ in range(2):
            # Do twice in case of "んーー"
            for c, r in zip(["んー", "ンー"], ["ん", "ン"]):
                word.pron = word.pron.replace(c, r)
                word.surface = word.surface.replace(c, r)

        for r in _REPLACEMENTS:
            word.pron = word.pron.replace(r, "")
            word.surface = word.surface.replace(r, "")
        word.pron = word.pron.replace(INTERNAL_SEP, "_")
        word.surface = word.surface.replace(INTERNAL_SEP, "_")
        # This is for pauses <P:00453.373-00454.013>
        word.pron = re.sub(r"<PL.+>", "", word.pron)

        # Occurs for example in noncore/A01F0063:
        # 0099 00280.998-00284.221 L:-001-001	一・	一・	イチ	一	イチ
        word.surface = word.surface.rstrip("・")

        # Make morph
        morph = [getattr(word, s) for s in _MORPH]
        word.morph = "/".join(m for m in morph if m)
        for c in ["Ａ", "１", "２", "３", "４"]:
            word.morph = word.morph.replace(c, "")
        word.morph = word.morph.replace("　", "＿")

        # Parse time
        word.sgid, start_end, channel = word.time.split(" ")
        word.start, word.end = [float(s) for s in start_end.split("-")]
        if word.spkid[0] == "D":
            word.spkid = word.spkid + "-" + channel.split(":")[0]

        return word

    def __repr__(self):
        return f"{self.surface}+{self.morph}+{self.pron}"

    def __bool__(self):
        return bool(self.surface or self.pron)


class _Transcript:
    text: str = ""
    shape0: List[int]
    shape1: List[int]
    shape2: List[int]
    tag_end: Dict

    def __init__(self, segments: List, text_type: str):
        self.shape0 = []
        self.shape1 = []
        self.shape2 = []
        self.tag_end = {}
        self.right_offset = defaultdict(list)
        for i, s in enumerate(segments):
            for j, w in enumerate(s):
                w = getattr(w, text_type)
                self.text += w
                for k in range(len(w)):
                    self.shape0.append(i)
                    self.shape1.append(j)
                    self.shape2.append(k)

        open_brackets = []
        for i, s in enumerate(self.text):
            if s == "(":
                open_brackets.append(i)
            elif s == ")":
                self.tag_end[open_brackets.pop()] = i

    def use_index(self, pos: int, right=False) -> Tuple[int]:
        if not right:
            return (self.shape0[pos], self.shape1[pos], self.shape2[pos])

        adjust = 0
        for coords in self.right_offset[(self.shape0[pos], self.shape1[pos])]:
            if coords < self.shape2[pos]:
                adjust += 1

        self.right_offset[(self.shape0[pos], self.shape1[pos])].append(self.shape2[pos])

        return (self.shape0[pos], self.shape1[pos], self.shape2[pos] - adjust)


class _CSJSDBSegment:
    text: str
    start: float
    end: float
    sgid: str

    @staticmethod
    def from_words(words: List[_CSJSDBWord]):
        ret = _CSJSDBSegment()
        ret.text = INTERNAL_SEP.join(str(w) for w in words)
        ret.start = words[0].start
        ret.end = words[-1].end
        ret.sgid = f"{words[0].spkid}_{words[0].sgid}"
        return ret

    def __repr__(self):
        return self.text

    def to_line(self):
        return f"{self.sgid}\t" + f"{self.start:09.3f}\t{self.end:09.3f}\t" + self.text

    def verify_line(self) -> bool:
        return self.text.count("(") == self.text.count(")")

    @staticmethod
    def from_line(line: str):
        ret = _CSJSDBSegment()
        ret.sgid, ret.start, ret.end, text = line.strip().split("\t")
        ret.start = float(ret.start)
        ret.end = float(ret.end)
        ret.text = text
        return ret


class CSJSDBParser:

    tag_regex = re.compile(r"( )|([\x00-\x7F])")
    JPN_NUM = [
        "ゼロ",
        "０",
        "零",
        "一",
        "二",
        "三",
        "四",
        "五",
        "六",
        "七",
        "八",
        "九",
        "十",
        "百",
        "千",
        "．",
    ]

    def __default_preprocess(self, x: str):
        ret = []
        for a in x.split(INTERNAL_SEP):
            a = a.split("+")[0]
            if a:
                ret.append(a)
        return INTERNAL_SEP.join(ret)

    def __init__(self, decisions: Dict = DECISIONS, preprocess: Callable = None):
        self.decisions = decisions
        if not preprocess:
            self.preprocess: Callable[[str], str] = self.__default_preprocess
        else:
            self.preprocess = preprocess

    def parse(
        self, text: str, sep="", with_tags=False
    ) -> Union[str, List[Tuple[str, ...]]]:
        processed_text = self.preprocess(text)
        ret = self._parse(processed_text, -1)

        assert len(ret["string"]) == len(ret["tag"]), text
        if not with_tags:
            return ret["string"].replace(INTERNAL_SEP, sep)

        if not sep:
            return [
                (w, t) for w, t in zip(ret["string"], ret["tag"]) if w != INTERNAL_SEP
            ]
        return [
            (w, t) if w != INTERNAL_SEP else (sep, t)
            for w, t in zip(ret["string"], ret["tag"])
        ]

    def _parse(self, text: str, open_bracket):
        i = open_bracket + 1
        tag = ""
        choices = [""]
        choices_tag = [[]]

        while i < len(text):
            c = text[i]
            t = [tag]

            if c == "(":
                ret = self._parse(text, i)
                c = ret["string"]
                i = ret["end"]
                if not tag:
                    t = ret["tag"]
                else:
                    t = [tag + f"/{ta}" for ta in ret["tag"]]

            matches = self.tag_regex.search(c)

            if c == ")" and not tag:
                print(f"{open_bracket}, {i}, {choices}, {tag}, {text}")
                return {"string": choices[-1], "end": i, "tag": choices_tag[-1]}
            elif c == ")":
                if tag == "A" and choices[0] and choices[0][0] in self.JPN_NUM:
                    tag = "A_num"

                result, result_tag = self._decide(
                    tag, choices + [""], choices_tag + [[]]
                )
                return {"string": result, "end": i, "tag": result_tag}

            elif c == ";":
                choices.append("")
                choices_tag.append([])
            elif c == ",":
                choices.append("")
                choices_tag.append([])
                if "," not in tag:
                    tag += ","
            elif c == "_":
                pass
            elif matches and matches.group(2):
                tag += c
            elif not tag and open_bracket > -1 and c in ["笑", "泣", "咳"]:
                tag = c
            else:
                choices[-1] += c
                choices_tag[-1].extend(t)
            i += 1
        return {
            "string": choices[-1],
            "end": i,
            "tag": choices_tag[-1] if choices[-1] else [],
        }

    def _decide(self, tag, choices, choices_tag) -> Tuple[str, List[str]]:
        assert len(choices) > 1

        for t, decision in self.decisions.items():
            if t != tag:
                continue
            ret = -1
            if isinstance(decision, int):
                ret = choices[decision], choices_tag[decision]
            elif decision[:5] == "eval:":
                ret = eval(decision[5:])
            elif decision[:5] == "exec:":
                exec(decision[5:])

            if ret != -1:
                return ret
            else:
                raise Exception(
                    f"Decision for {tag} cannot be resolved. Got {decision}"
                )

        raise NotImplementedError(f"Unknown tag {tag} encountered.")


class _CSJSDBTagSegment:
    segments: List[List[_CSJSDBWord]]
    surface_open_brackets: Dict
    pron_open_brackets: Dict

    def __init__(self):
        self.segments = []
        self.surface_open_brackets = {}
        self.pron_open_brackets = {}

    def append(self, word: _CSJSDBWord):
        if self.segments:
            self.segments[-1].append(word)
        else:
            self.segments = [[word]]

    def flatten(self) -> _CSJSDBSegment:
        return _CSJSDBSegment.from_words(words=[w for s in self.segments for w in s])

    def split(self) -> List[_CSJSDBSegment]:
        return [_CSJSDBSegment.from_words(words=s) for s in self.segments if s]

    def __getitem__(self, pos):
        return self.segments[pos]

    @property
    def is_complete(self):
        surface = "".join(w.surface for s in self.segments for w in s)
        surface_open_brackets = self._get_open_brackets(surface)
        pron = "".join(w.pron for s in self.segments for w in s)
        pron_open_brackets = self._get_open_brackets(pron)

        if not surface_open_brackets and not pron_open_brackets:
            return True

        self.surface_open_brackets.update(
            {ii: surface[ii + 1] for ii in surface_open_brackets[::-1]}
        )

        self.pron_open_brackets.update(
            {ii: pron[ii + 1] for ii in pron_open_brackets[::-1]}
        )
        return False

    def _get_open_brackets(self, text) -> List[int]:
        brackets = []
        for i, s in enumerate(text):
            if s == "(":
                brackets.append(i)
            elif s == ")":
                brackets.pop()

        return brackets

    def __bool__(self):
        return bool(self.segments and self.segments[0])


def _read_one_sdb(sdb: Path) -> List[_CSJSDBSegment]:
    lines = sdb.read_text(encoding="shift_jis").split("\n")
    sgid = lines[0].split("\t")[3].split(" ")[0]

    words_until_now = _CSJSDBTagSegment()
    segments: List[_CSJSDBSegment] = []

    for line in lines:
        if not line:
            # Last line is always empty
            word = _CSJSDBWord()
        else:
            word = _CSJSDBWord.from_line(line)

        if not word and line:
            continue

        if word.sgid == sgid:
            words_until_now.append(word)
            continue

        sgid = word.sgid

        if not words_until_now.is_complete:
            words_until_now.segments.append([])
            pass
        elif not words_until_now:
            pass
        elif len(words_until_now.segments) > 1:
            surface = _Transcript(words_until_now, "surface")
            pron = _Transcript(words_until_now, "pron")

            for pos, linking_tag in words_until_now.pron_open_brackets.items():
                if linking_tag in ["R", "M", "O"]:
                    l0, l1, l2 = pron.use_index(pos)
                    r0, r1, r2 = pron.use_index(pron.tag_end[pos], True)
                    # r2 -= i
                    l1_p = words_until_now[l0][l1].pron
                    r1_p = words_until_now[r0][r1].pron
                    words_until_now[l0][l1].pron = l1_p[:l2] + l1_p[l2 + 3 :]
                    words_until_now[r0][r1].pron = r1_p[:r2] + r1_p[r2 + 1 :]

            split = True
            for pos, linking_tag in words_until_now.surface_open_brackets.items():
                if linking_tag in ["R", "M", "O"]:
                    l0, l1, l2 = surface.use_index(pos)
                    r0, r1, r2 = surface.use_index(surface.tag_end[pos], True)
                    # r2 -= i
                    l1_s = words_until_now[l0][l1].surface
                    r1_s = words_until_now[r0][r1].surface
                    words_until_now[l0][l1].surface = l1_s[:l2] + l1_s[l2 + 3 :]
                    words_until_now[r0][r1].surface = r1_s[:r2] + r1_s[r2 + 1 :]
                else:
                    split = False

            if split:
                segments.extend(words_until_now.split())
            else:
                segments.append(words_until_now.flatten())
            words_until_now = _CSJSDBTagSegment()
        else:
            segments.append(words_until_now.flatten())
            words_until_now = _CSJSDBTagSegment()

        words_until_now.append(word)

    return segments


def _process_one_recording(
    segments: List[_CSJSDBSegment],
    wav: Path,
    recording_id: str,
    parser: CSJSDBParser,
) -> Tuple[Recording, List[SupervisionSegment]]:
    recording = Recording.from_file(wav, recording_id=recording_id)

    supervision_segments = []

    for segment in segments:
        text = parser.parse(segment.text, sep="", with_tags=True)
        if not text:
            continue
        text, tag = list(zip(*text))
        text = "".join(text)
        tag = ",".join(tag)
        supervision_segments.append(
            SupervisionSegment(
                id=segment.sgid,
                recording_id=recording_id,
                start=segment.start,
                duration=(segment.end - segment.start),
                channel=0,
                language="Japanese",
                speaker=recording_id,
                gender=("Male" if recording_id[3] == "M" else "Female"),
                text=text,
                custom={"raw": segment.text, "disfluent": text, "disfluent_tag": tag},
            )
        )
    return recording, supervision_segments


def _process_one(sdb: Path, parser: CSJSDBParser):
    segments = _read_one_sdb(sdb)
    spk = sdb.stem
    try:
        wavfile = Path((sdb.parent / (spk + "-wav.list")).read_text())
        (sdb.parent / f"{sdb.stem}-trans.txt").write_text(
            "\n".join(s.to_line() for s in segments)
        )
    except FileNotFoundError:
        part = sdb.parent.name
        wavfile = sdb.parents[3] / (f"WAV/{part}/{spk}.wav")
        assert wavfile.exists()
    return _process_one_recording(segments, wavfile, spk, parser)


def prepare_manifests(
    transcript_dir: Path,
    dataset_parts: Union[str, Sequence[str]] = None,
    manifest_dir: Pathlike = None,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will
    simply read and return them.

    :param transcript_dir: Path, the path to the .sdb transcripts.
    :param dataset_parts: string or sequence of strings representing
        dataset part names, e.g. 'eval1', 'core', 'eval2'. This defaults to the
        full dataset - core, noncore, eval1, eval2, and eval3.
    :param manifest_dir: Pathlike, the path where to write the manifests.
    :param num_jobs: int, to be passed to ThreadPoolExecutor.
    :return: a Dict whose key is the dataset part, and the value is Dicts
        with the keys 'recordings' and 'supervisions'.
    """
    assert (
        transcript_dir.is_dir()
    ), f"No such directory for transcript_dir: {transcript_dir}"
    if not dataset_parts:
        dataset_parts = _FULL_DATA_PARTS
    elif isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]

    glob_pattern = "*.sdb" if transcript_dir.name == "SDB" else "*/*.sdb"

    manifests = {}

    if manifest_dir:
        manifest_dir = Path(manifest_dir)
        manifest_dir.mkdir(parents=True, exist_ok=True)
        # Maybe the manifests already exit: we can read them and
        # save a bit of preparation time.
        manifests = read_manifests_if_cached(
            dataset_parts=dataset_parts,
            output_dir=manifest_dir,
            prefix="csj",
        )
    parser = CSJSDBParser(DECISIONS)
    with ThreadPoolExecutor(num_jobs) as ex:
        for part in tqdm(dataset_parts, desc="Dataset parts"):
            futures = []
            logging.info(f"Processing CSJ subset: {part}")
            if manifests_exist(part=part, output_dir=manifest_dir, prefix="csj"):
                logging.info(f"CSJ subset: {part} already prepared - skipping.")
                continue
            for sdb in transcript_dir.glob(f"{part}/{glob_pattern}"):
                futures.append(ex.submit(_process_one, sdb, parser))

            recordings = []
            supervisions = []
            for future in tqdm(futures, desc="Processing", leave=False):
                result: Tuple[Recording, List[SupervisionSegment]] = future.result()
                assert result
                recording, segments = result
                recordings.append(recording)
                supervisions.extend(segments)

            recording_set = RecordingSet.from_recordings(recordings)
            supervision_set = SupervisionSet.from_segments(supervisions)

            recording_set, supervision_set = fix_manifests(
                recording_set, supervision_set
            )
            validate_recordings_and_supervisions(recording_set, supervision_set)

            if manifest_dir:
                supervision_set.to_file(
                    manifest_dir / f"csj_supervisions_{part}.jsonl.gz"
                )
                recording_set.to_file(manifest_dir / f"csj_recordings_{part}.jsonl.gz")

            manifests[part] = {
                "recordings": recording_set,
                "supervisions": supervision_set,
            }
    return manifests


def prepare_csj(
    corpus_dir: Pathlike,
    transcript_dir: Pathlike = None,
    manifest_dir: Pathlike = None,
    dataset_parts: Union[str, Sequence[str]] = None,
    nj: int = 16,
):
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir()
    if transcript_dir:
        transcript_dir = Path(transcript_dir)
        transcript_dir.mkdir(parents=True, exist_ok=True)
        logging.info("Creating transcript directories now.")
        _create_trans_dir(corpus_dir, transcript_dir)
    else:
        transcript_dir = corpus_dir / "MORPH" / "SDB"
        logging.info(
            "Preparing manifests without saving transcripts. Only core and "
            "noncore can be created. "
        )
        if not dataset_parts:
            dataset_parts = ["core", "noncore"]
    return prepare_manifests(
        transcript_dir=transcript_dir,
        dataset_parts=dataset_parts,
        manifest_dir=manifest_dir,
        num_jobs=nj,
    )


def concat_csj_supervisions(
    supervisions: SupervisionSet,
    gap: float,
    maxlen: float,
    max_extend_right=0.0,
) -> SupervisionSet:
    """Concatenates supervisions according to the permissible gap and maximum length
    of an utterance. This function is not called in this script, but is provided as
    a utility function for users to concatenate supervisions themselves.

    Args:
        supervisions (SupervisionSet): The list of `SupervisionSegments` to concatenate.
        gap (float): Maximum length of permissible silence in an utterance.
        maxlen (float): Maximum length of one utterance.
        max_extend_right (float, optional): Since the immediate right context is silence,
        optionally extends the supervision to include some silence on the right. Included
        for endpointing training. Defaults to 0.0.

    Returns:
        SupervisionSet: Concatenated supervisions
    """
    grouped_supervisions = []
    supervisions = copy.deepcopy(supervisions)
    tmp_sp = []
    for long_sp0 in supervisions:
        if "×" in long_sp0.custom["raw"]:
            if tmp_sp:
                grouped_supervisions.append(tmp_sp)
                tmp_sp = []
        elif not tmp_sp:
            tmp_sp.append(long_sp0)
        elif long_sp0.speaker != tmp_sp[0].speaker:
            grouped_supervisions.append(tmp_sp)
            tmp_sp = [long_sp0]
        elif (long_sp0.end - tmp_sp[0].start) >= maxlen:
            grouped_supervisions.append(tmp_sp)
            tmp_sp = [long_sp0]
        elif (long_sp0.start - tmp_sp[-1].end) >= gap:
            tmp_sp[-1].duration += min(
                max_extend_right, long_sp0.start - tmp_sp[-1].end
            )
            grouped_supervisions.append(tmp_sp)
            tmp_sp = [long_sp0]
        else:
            tmp_sp.append(long_sp0)
    if tmp_sp:
        grouped_supervisions.append(tmp_sp)
    ret = []
    for long_sp in grouped_supervisions:
        long_sp0 = long_sp[0]
        long_sp0.duration = long_sp[-1].end - long_sp0.start
        for k in long_sp0.custom:
            if k == "raw":
                long_sp0.custom[k] = " ".join(sp.custom[k] for sp in long_sp)
            elif "_tag" in k:
                long_sp0.custom[k] = ",".join(sp.custom[k] for sp in long_sp)
            else:
                long_sp0.custom[k] = "".join(sp.custom[k] for sp in long_sp)
        long_sp0.text = "".join(sp.text for sp in long_sp)

        ret.append(long_sp0)

    return SupervisionSet.from_segments(ret)



"""
About the HI_MIA corpus
HI_MIA is a far-field text-dependent speaker verification data
published by Beijing Shell Shell Technology Co.,Ltd.
The contents are wake-up words "Hi, Mia" in Chinese(ni hao mi ya; 你好，米雅).
It' publicly available on https://www.openslr.org/85

The HI_MIA_CW is a supplemental database of the HI_MIA database.
The specific text of the audios is the HI-MIA confusion words in Chinese,
which are the negative samples for wake-up words "hi, Mia"(ni hao mi ya; 你好, 米雅).
It' publicly available on https://www.openslr.org/120
"""

import logging
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download, safe_extract

# HI_MIA contains train.tar.gz, dev.tar.gz, test.tar.gz and test_v2.tar.gz
# According to https://www.openslr.org/85,
# test.tar.gz is deprecated because of corrupted audio files.
# So here we download test_v2.tar.gz instead of test.tar.gz
# Following three files are going to be downloaded:
#   https://us.openslr.org/resources/85/train.tar.gz
#   https://us.openslr.org/resources/85/dev.tar.gz
#   https://us.openslr.org/resources/85/test_v2.tar.gz
#
# The extracted folder of test_v2.tar.gz is "test" rather than "test_v2".

# HI_MIA_CW contains data.tgz and resource.tgz
# https://us.openslr.org/resources/120/data.tgz
# will be extracted to folder "16k_wav_file"
#
# Transcriptions.
# https://us.openslr.org/resources/120/resource.tgz
# will be extracted to folder "resource"

SOURCE_FILE = {
    "train": "train.tar.gz",
    "dev": "dev.tar.gz",
    "test": "test_v2.tar.gz",
    # "cw_test" contains following two entries.
    "data": "data.tgz",
    "resource": "resource.tgz",
}

EXTRACTED_FOLDER = {
    "train": "train",
    "dev": "dev",
    "test": "test",
    # "cw_test" contains following two entries.
    "data": "16k_wav_file",
    "resource": "resource",
}

CW_PARTS = ["cw_test"]
CW_SOURCE_FILE_LIST = ["data", "resource"]
CW_FILES = ["data.tgz", "resource.tgz"]

HI_MIA_PARTS = ["train", "dev", "test"]
HI_MIA_AND_CW_PARTS = HI_MIA_PARTS + CW_PARTS


def _validate_dataset_parts(
    dataset_parts: Optional[Union[str, Sequence[str]]] = "auto",
) -> bool:
    valid_dataset_parts = HI_MIA_AND_CW_PARTS + ["auto", "himia"]

    def validate_a_dataset(dataset_name: str) -> bool:
        assert dataset_name in valid_dataset_parts, (
            f"{dataset_name} is not a valid subset. "
            f"You may want to select one from {valid_dataset_parts}"
        )
        return True

    if isinstance(dataset_parts, str):
        validate_a_dataset(dataset_parts)
        return True
    assert isinstance(dataset_parts, tuple)
    for dataset_name in dataset_parts:
        validate_a_dataset(dataset_name)
    return True


def download_himia(
    target_dir: Pathlike = ".",
    dataset_parts: Optional[Union[str, Sequence[str]]] = "auto",
    force_download: bool = False,
    base_url: str = "http://www.openslr.org/resources",
) -> Path:
    """
    Downdload and untar HI_MIA and HI_MIA_CW datasets.
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: "auto", "himia"
        or a list of splits (e.g. "train", "dev", "test", "cw_test") to download.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    :return: the path to extracted directory with data.
    """
    target_dir = Path(target_dir)
    corpus_dir = target_dir / "HiMia"

    _validate_dataset_parts(dataset_parts)
    if dataset_parts == "auto":
        dataset_parts = HI_MIA_PARTS + CW_SOURCE_FILE_LIST
    elif dataset_parts == "himia":
        dataset_parts = HI_MIA_PARTS
    elif isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]

    files_to_download = []
    # Example: when command with `-p test -p cw_test`
    # All previous if/elif/elif branches are not executed.
    # We need following loop to expand "cw_test" to CW_SOURCE_FILE_LIST
    for dataset_name in dataset_parts:
        if dataset_name == "cw_test":
            files_to_download += CW_SOURCE_FILE_LIST
        else:
            files_to_download.append(dataset_name)

    tar_files = [SOURCE_FILE[part] for part in files_to_download]
    ext_folders = [EXTRACTED_FOLDER[part] for part in files_to_download]

    for tar_name, ext_name in zip(tar_files, ext_folders):
        # HI_MIA_CW is from
        # https://us.openslr.org/resources/120/
        # HI_MIA is from
        # https://us.openslr.org/resources/85/
        is_cw = True if tar_name in CW_FILES else False
        url_suffix_index = 120 if is_cw else 85
        url = f"{base_url}/{url_suffix_index}"

        tar_path = target_dir / tar_name
        completed_detector_dir = corpus_dir / ext_name
        if is_cw:
            completed_detector_dir = corpus_dir / "cw_test" / ext_name
        completed_detector = completed_detector_dir / ".completed"

        extracted_dir = (
            completed_detector_dir
            if "resource.tgz" == tar_name
            else completed_detector_dir.parent
        )

        if completed_detector.is_file():
            logging.info(
                f"Skipping download and extraction of {tar_name} because {completed_detector} exists."
            )
            continue
        resumable_download(
            f"{url}/{tar_name}", filename=tar_path, force_download=force_download
        )

        logging.info(f"Extracting {tar_name}.")
        shutil.rmtree(completed_detector_dir, ignore_errors=True)
        extracted_dir.mkdir(parents=True, exist_ok=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=extracted_dir)
        completed_detector.touch()

    return corpus_dir


_TOTAL_NUM_WAVS = {"train": 993083, "dev": 164640, "test": 165120, "cw_test": 16343}


def _prepare_train_dev_test(
    corpus_dir: Path,
    part: str,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param corpus_dir: Pathlike, the path of the data dir.
    :param part: dataset part, one of ["train", "dev", "test"]
    :return: the RecodingSet and SupervisionSet given a dataset part.
    """
    logging.info(f"Processing HI_MIA subset: {part}")
    suffix_path = "" if part == "test" else "SPEECHDATA"
    scp_file_name = "wav" if part == "test" else part
    dir_of_wav_scp = corpus_dir / f"{part}/{suffix_path}/"
    wav_scp_path = dir_of_wav_scp / f"{scp_file_name}.scp"
    assert wav_scp_path.is_file(), f"{wav_scp_path}"
    recordings = []
    supervisions = []

    wav_path = "wav/" if part == "test" else ""
    with open(wav_scp_path) as wav_scp_f:
        for wav_entry in tqdm(wav_scp_f, total=_TOTAL_NUM_WAVS[part]):
            wav_entry = wav_entry.strip()
            audio_path = dir_of_wav_scp / f"{wav_path}" / f"{wav_entry}"
            audio_path = audio_path.resolve()
            audio_file_name = audio_path.stem
            # Example of audio_file_name: SV0297_2_00_F0041
            # speaker: SV0297
            # See rule of the filename at: https://aishelltech.com/wakeup_data
            speaker = audio_file_name.split("_")[0]
            # According to https://www.aishelltech.com/wakeup_data
            # The wake-up word is "你好，米雅".
            # The comma is removed to save space.
            text = "你好米雅"
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            segment = SupervisionSegment(
                id=audio_file_name,
                recording_id=audio_file_name,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="Chinese",
                speaker=speaker,
                text=text.strip(),
            )
            supervisions.append(segment)
    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)
    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)
    return recording_set, supervision_set


def _prepare_cw_test(corpus_path: Path) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet of test dataset.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet of test dataset.
    """
    logging.info("Processing HI_MIA_CW dataset")
    recordings = []
    supervisions = []
    cw_test_path = corpus_path / "cw_test/16k_wav_file"
    logging.info(f"Searching wav files in {cw_test_path}")
    transcript_path = corpus_path / "cw_test/resource/transcription.txt"
    transcript_dict = {}
    with open(transcript_path, "r", encoding="utf-8") as f:
        for line in f.readlines():
            # Examples of line:
            # 0001_M02_01_fast_0001.wav▸你好米
            # 0001_M02_01_fast_0002.wav▸你好你好
            # 0001_M02_01_fast_0003.wav▸你好亚

            idx_transcript = line.split()
            transcript = " ".join(idx_transcript[1:])
            transcript_dict[idx_transcript[0]] = transcript

    assert len(transcript_dict) == _TOTAL_NUM_WAVS["cw_test"]

    for wav_name in tqdm(transcript_dict, total=_TOTAL_NUM_WAVS["cw_test"]):
        audio_path = cw_test_path / wav_name
        audio_path = audio_path.resolve()
        text = transcript_dict[wav_name]
        assert audio_path.is_file(), f"{audio_path} does not exist."
        recording = Recording.from_file(audio_path)
        recordings.append(recording)
        audio_file_name = audio_path.stem
        speaker = audio_file_name.split("_")[0]
        segment = SupervisionSegment(
            id=audio_file_name,
            recording_id=audio_file_name,
            start=0.0,
            duration=recording.duration,
            channel=0,
            language="Chinese",
            speaker=speaker,
            text=text.strip(),
        )
        supervisions.append(segment)

    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)
    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)
    return recording_set, supervision_set


def prepare_himia(
    corpus_dir: Pathlike,
    dataset_parts: Union[str, Sequence[str]] = "auto",
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param dataset_parts: "auto", "himia"
        or a list of splits (e.g. "train", "dev", "test", "cw_test") to download.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    _validate_dataset_parts(dataset_parts)
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    manifests = defaultdict(dict)
    if dataset_parts == "auto":
        dataset_parts = HI_MIA_AND_CW_PARTS
    elif dataset_parts == "himia":
        dataset_parts = HI_MIA_PARTS
    elif isinstance(dataset_parts, str):
        if dataset_parts == "cw_test":
            dataset_parts = CW_PARTS
        else:
            dataset_parts = [dataset_parts]

    for part in tqdm(
        dataset_parts,
        desc="Process HI_MIA and HI_MIA_CW dataset.",
    ):
        if "cw_test" == part:
            recording_set, supervision_set = _prepare_cw_test(corpus_dir)
        else:
            recording_set, supervision_set = _prepare_train_dev_test(corpus_dir, part)
        if output_dir is not None:
            supervision_set.to_file(output_dir / f"himia_supervisions_{part}.jsonl.gz")
            recording_set.to_file(output_dir / f"himia_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



# Copyright    2023  Johns Hopkins        (authors: Amir Hussein, Matthew Wiesner)

"""
The IWSLT Tunisian dataset is a 3-way parallel dataset consisting of approximately 160 hours
and 200,000 lines of aligned audio, Tunisian transcripts, and English translations. This dataset
comprises conversational telephone speech recorded at a sampling rate of 8kHz. The train, dev,
and test1 splits of the iwslt2022 shared task correspond to catalog number LDC2022E01. Please
note that access to this data requires an LDC subscription from your institution.To obtain this
dataset, you should download the predefined splits by running the following command:
git clone https://github.com/kevinduh/iwslt22-dialect.git. For more detailed information about
the shared task, please refer to the task paper available at this link:
https://aclanthology.org/2022.iwslt-1.10/.
"""

import logging
import re
import string
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import tqdm

from lhotse import (
    AudioSource,
    Recording,
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    validate_recordings_and_supervisions,
)
from lhotse.qa import fix_manifests
from lhotse.utils import Pathlike

# English annotation rules:
# -------------------------------
#     (()) - Uncertain word or words
#      %pw - Partial word
#       #  - Foreign word, either followed by translation or (()) if cannot translate
#       +  - Mis-pronounced word (carried over from mispronunciation marked in transcript)
#      uh, um, eh or ah - Filled pauses
#       =  - Typographical error from transcript


# Arabic annotation rules:
# -------------------------------
# O/ - foreign
# U/ - uncertain
# M/ - MSA
# UM/ - uncertain + MSA
# UO/ - uncertain + foreign


def download_iwslt22_ta(
    target_dir: Pathlike = ".",
) -> None:
    """
    Download and untar the dataset.

    NOTE: This function just returns with a message since IWSLT22 Tunisian-English is not available
    for direct download.

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    """
    logging.info(
        """
        To obtain this data your institution needs to have an LDC subscription.
        You also should download the pre-defined splits with
        git clone https://github.com/kevinduh/iwslt22-dialect.git
    """
    )


def prepare_iwslt22_ta(
    corpus_dir: Pathlike,
    splits: Pathlike,
    output_dir: Optional[Pathlike] = None,
    normalize_text: bool = False,
    langs: Optional[List[str]] = ["ta", "eng"],
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Prepares manifests for the train dev and test1 splits.

    :param corpus_dir: Path to ``LDC2022E01`` the path of the data dir.
    :param splits: Path to splits from https://github.com/kevinduh/iwslt22-dialect
    :param normalize_text: Bool, if True, Arabic text normalization is performed
        from https://aclanthology.org/2022.iwslt-1.29.pdf.
    :param output_dir: Directory where the manifests should be written. Can be omitted
        to avoid writing.
    :param langs: str, list of language abbreviations for source and target languages.
    :param num_jobs: int, the number of jobs to use for parallel processing.
    :return: A dict with manifests. The keys are: ``{'recordings', 'supervisions'}``.

    """
    import soundfile as sf

    manifests = {}
    split_files = load_splits(Path(splits))
    corpus_dir = Path(corpus_dir)
    audio_dir = corpus_dir / "data/audio/ta"
    exclude = []
    with open(str(Path(splits) / "exclude-utterance.txt")) as f:
        for l in f:
            excludeid, start, end = l.strip().split()
            start = float(start)
            exclude.append(f"{excludeid}_{int(100*start):06}")

    recordings = {}
    supervisions = []
    text_dir = corpus_dir / "data/transcripts/ta"
    futures = []
    with ThreadPoolExecutor(num_jobs) as ex:
        for p in text_dir.glob("*.tsv"):
            if not p.stem.startswith("._"):
                translations_path = (
                    p.parent.parent.parent / "translations" / "ta" / p.name
                )
                translations_path = translations_path.with_name(
                    p.stem.split(".")[0] + ".eng" + p.suffix
                )
                if translations_path.exists():
                    filename = p.with_suffix("").stem
                    audio_sf = sf.SoundFile(str(audio_dir / f"{filename}.sph"))
                    if filename not in recordings:
                        recordings[filename] = Recording(
                            id=filename,
                            sources=[
                                AudioSource(
                                    type="file",
                                    channels=[0],
                                    source=str(audio_dir / f"{filename}.sph"),
                                ),
                            ],
                            sampling_rate=audio_sf.samplerate,
                            num_samples=audio_sf.frames,
                            duration=audio_sf.frames / audio_sf.samplerate,
                        )
                    futures.append(
                        ex.submit(
                            _filename_to_supervisions,
                            p,
                            translations_path,
                            normalize_text,
                            exclude,
                            langs,
                        )
                    )
                else:
                    logging.warning(
                        f"{translations_path.stem} does not exist, please make sure number of translations = transcriptions"
                    )

        for future in tqdm.tqdm(futures, desc="Processing text", leave=False):
            result = future.result()
            if result is None:
                continue
            for sup in result:
                supervisions.append(sup)

        supervisions = deduplicate_supervisions(supervisions)
        supervisions = SupervisionSet.from_segments(supervisions)
        recordings = RecordingSet.from_recordings(recordings.values())
        recordings, supervisions = fix_manifests(recordings, supervisions)
        validate_recordings_and_supervisions(recordings, supervisions)
        for split in ("train", "dev", "test1"):
            sups_ = supervisions.filter(lambda s: s.recording_id in split_files[split])
            recs_ = recordings.filter(lambda r: r.id in split_files[split])
            manifests[split] = {"recordings": recs_, "supervisions": sups_}

            if output_dir is not None:
                output_dir = Path(output_dir)
                output_dir.mkdir(parents=True, exist_ok=True)
                sups_.to_file(output_dir / f"iwslt22-ta_recordings_{split}.jsonl.gz")
                recs_.to_file(output_dir / f"iwslt22-ta_supervisions_{split}.jsonl.gz")

    return manifests


def _filename_to_supervisions(
    p: Path,
    translations_path: Path,
    normalize_text: bool,
    exclude: list,
    langs: list,
):
    supervisions = []
    date, time, someid, channel = p.with_suffix("").stem.split("_")
    date_tgt, time_tgt, someid_tgt, channel_tgt = p.with_suffix("").stem.split("_")
    translations = translations_path.read_text().splitlines()
    transcripts = p.read_text().splitlines()

    sorted_translations = sorted(translations, key=lambda line: line.split("\t")[0])
    sorted_transcripts = sorted(transcripts, key=lambda line: line.split("\t")[0])

    for src, tgt in zip(sorted_transcripts, sorted_translations):

        start, end, sid, text = src.rstrip().split("\t")
        _, _, _, text_tgt = tgt.rstrip().split("\t")
        start = float(start)
        end = float(end)

        # Following the IWSLT provided text normalization
        text = normalize_text(text, "transcript")
        text_tgt = normalize_text(text_tgt, "translation")

        utt_id = f"{date}_{time}_{someid}_{channel}_{int(100*start):06}"
        utt_id_tgt = (
            f"{date_tgt}_{time_tgt}_{someid_tgt}_{channel_tgt}_{int(100*start):06}"
        )

        assert (
            utt_id == utt_id_tgt
        ), f"The loaded source and target files are not sorted properly: {utt_id} {utt_id_tgt}"

        if normalize_text:
            # Aggressive Tunisian text normalization from https://aclanthology.org/2022.iwslt-1.29.pdf
            text = text_cleaning(text)
            if text.strip() == "":
                logging.warning(
                    f"Skipping {p.stem} {start} {end} {text} with empty cleaned transcript ..."
                )
                continue

        if utt_id in exclude:
            continue
        supervisions.append(
            SupervisionSegment(
                id=f"{sid}_{langs[0]}_{langs[1]}_{utt_id}",
                recording_id=p.with_suffix("").stem,
                start=start,
                duration=round(end - start, ndigits=8),
                channel=0,
                text=text,
                language=langs[0],
                speaker=sid,
                custom={"translated_text": {langs[1]: text_tgt}},
            )
        )
    return supervisions


def normalize_text(utterance, language):
    arabic_filter = re.compile(r"[OUM]+/*|\u061F|\?|\!|\.")
    english_filter = re.compile(r"\(|\)|\#|\+|\=|\?|\!|\;|\.|\,|\"|\:")
    if language == "transcript":
        return re.subn(arabic_filter, "", utterance)[0]
    elif language == "translation":
        return re.subn(english_filter, "", utterance)[0].lower()
    else:
        raise ValueError(f"Text normalization for {language} is not supported")


def load_splits(path):
    splits = {}
    for split in ("train", "dev", "test1"):
        file_list = []
        split_scp = path / f"{split}.file_id.txt"
        with open(str(split_scp)) as f:
            for l in f:
                file_list.append(l.strip())
        splits[split] = file_list
    return splits


def deduplicate_supervisions(
    supervisions: Iterable[SupervisionSegment],
) -> List[SupervisionSegment]:
    from cytoolz import groupby

    duplicates = groupby((lambda s: s.id), sorted(supervisions, key=lambda s: s.id))
    filtered = []
    for k, v in duplicates.items():
        if len(v) > 1:
            logging.warning(
                f"Found {len(v)} supervisions with conflicting IDs ({v[0].id}) "
                f"- keeping only the first one."
            )
        filtered.append(v[0])
    return filtered


_preNormalize = " \u0629\u0649\u0623\u0625\u0622"
_postNormalize = " \u0647\u064a\u0627\u0627\u0627"
_toNormalize = {ord(b): a for a, b in zip(_postNormalize, _preNormalize)}


def normalize_text_(s):
    return s.translate(_toNormalize)


def normalize_arabic(text):
    text = re.sub("[إأٱآا]", "ا", text)
    text = re.sub(r"(أ){2,}", "ا", text)
    text = re.sub(r"(ا){2,}", "ا", text)
    text = re.sub(r"(آ){2,}", "ا", text)
    text = re.sub(r"(ص){2,}", "ص", text)
    text = re.sub(r"(و){2,}", "و", text)
    return text


def remove_diacritics(text):
    # https://unicode-table.com/en/blocks/arabic/
    return re.sub(r"[\u064B-\u0652\u06D4\u0670\u0674\u06D5-\u06ED]+", "", text)


def remove_punctuations(text):
    """This function  removes all punctuations except the verbatim"""

    arabic_punctuations = """`÷×؛<>_()*&^%][ـ،/:"؟.,'{}~¦+|!”…“–ـ"""
    english_punctuations = string.punctuation
    all_punctuations = set(
        arabic_punctuations + english_punctuations
    )  # remove all non verbatim punctuations

    for p in all_punctuations:
        if p in text:
            text = text.replace(p, " ")
    return text


def remove_extra_space(text):
    text = re.sub("\s+", " ", text)
    text = re.sub("\s+\.\s+", ".", text)
    return text


def east_to_west_num(text):
    eastern_to_western = {
        "٠": "0",
        "١": "1",
        "٢": "2",
        "٣": "3",
        "٤": "4",
        "٥": "5",
        "٦": "6",
        "٧": "7",
        "٨": "8",
        "٩": "9",
        "٪": "%",
        "_": " ",
        "ڤ": "ف",
        "|": " ",
    }
    trans_string = str.maketrans(eastern_to_western)
    return text.translate(trans_string)


def text_cleaning(text):
    text = remove_punctuations(text)
    text = east_to_west_num(text)
    text = remove_diacritics(text)
    text = remove_extra_space(text)
    text = normalize_arabic(text)
    text = normalize_text_(text)
    return text



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.dipco import download_dipco, prepare_dipco
from lhotse.utils import Pathlike

__all__ = ["dipco"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--mic",
    type=click.Choice(["ihm", "mdm"], case_sensitive=False),
    default="mdm",
    help="DiPCo microphone setting.",
)
@click.option(
    "--normalize-text",
    type=click.Choice(["none", "upper", "kaldi"], case_sensitive=False),
    default="kaldi",
    help="Text normalization method.",
    show_default=True,
)
@click.option(
    "--use-chime7-offset",
    is_flag=True,
    default=False,
    help="If True, offset session IDs (from CHiME-7 challenge).",
)
def dipco(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    mic: str,
    normalize_text: str,
    use_chime7_offset: bool,
):
    """DiPCo data preparation."""
    prepare_dipco(
        corpus_dir,
        output_dir=output_dir,
        mic=mic,
        normalize_text=normalize_text,
        use_chime7_offset=use_chime7_offset,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "--force-download",
    type=bool,
    default=False,
    help="If True, download even if file is present.",
)
def dipco(
    target_dir: Pathlike,
    force_download: bool,
):
    """DiPCo download."""
    download_dipco(
        target_dir,
        force_download=force_download,
    )



"""
The following are the original TED-LIUM 2 README contents.

This is theTED-LIUM corpus release 2, English speech recognition training corpus from TED talks, created by Laboratoire d’Informatique de l’Université du Maine (LIUM) (mirrored here)
licensed under Creative Commons BY-NC-ND 3.0 (http://creativecommons.org/licenses/by-nc-nd/3.0/deed.en).

All talks and text are property of TED Conferences LLC.

--- 

The TED-LIUM corpus was made from audio talks and their transcriptions available on the TED website. We have prepared and filtered these data in order to train acoustic models to participate to the International Workshop on Spoken Language Translation 2011 (the LIUM English/French SLT system reached the first rank in the SLT task). 

More details are given in this paper: 

A. Rousseau, P. Deléglise, and Y. Estève, "Enhancing the TED-LIUM Corpus with Selected Data for Language Modeling and More TED Talks",
in Proceedings of the Ninth International Conference on Language Resources and Evaluation (LREC’14), May 2014.


Please cite this reference if you use these data in your research work. 

--- 

Contents: 

- 1495 audio talks in NIST sphere format (SPH) 
- 1495 transcripts in STM format 
- Dictionary with pronunciation (159848 entries) 
- Selected monolingual data for language modeling from WMT12 publicly available corpora


SPH format info: 

Channels			: 1
Sample Rate		: 16000
Precision			: 16-bit
Bit Rate			: 256k
Sample Encoding	: 16-bit Signed Integer PCM

"""

import logging
import shutil
import tarfile
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from lhotse import (
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    validate_recordings_and_supervisions,
)
from lhotse.qa import fix_manifests
from lhotse.recipes.utils import normalize_text_tedlium
from lhotse.utils import Pathlike, resumable_download, safe_extract

TEDLIUM_PARTS = ("train", "dev", "test")


def download_tedlium2(
    target_dir: Pathlike = ".", force_download: Optional[bool] = False
) -> Path:
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    tar_path = target_dir / "TEDLIUM_release2.tar.gz"
    corpus_dir = target_dir / "TEDLIUM_release2"
    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {tar_path.name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(
        "http://www.openslr.org/resources/19/TEDLIUM_release2.tar.gz",
        filename=tar_path,
        force_download=force_download,
    )
    shutil.rmtree(corpus_dir, ignore_errors=True)
    with tarfile.open(tar_path) as tar:
        safe_extract(tar, path=target_dir)
    completed_detector.touch()
    return corpus_dir


def prepare_tedlium2(
    tedlium_root: Pathlike,
    output_dir: Optional[Pathlike] = None,
    dataset_parts: Union[str, Sequence[str]] = TEDLIUM_PARTS,
    num_jobs: int = 1,
    normalize_text: str = "none",
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Prepare manifests for the TED-LIUM v2 corpus.

    The manifests are created in a dict with three splits: train, dev and test.
    Each split contains a RecordingSet and SupervisionSet in a dict under keys 'recordings' and 'supervisions'.

    :param tedlium_root: Path to the unpacked TED-LIUM data.
    :param output_dir: Path where the manifests should be written.
    :param dataset_parts: Which parts of the dataset to prepare.
        By default, all parts are prepared.
    :param num_jobs: Number of parallel jobs to use.
    :return: A dict with standard corpus splits containing the manifests.
    """
    tedlium_root = Path(tedlium_root)
    output_dir = Path(output_dir) if output_dir is not None else None
    corpus = {}

    dataset_parts = [dataset_parts] if isinstance(dataset_parts, str) else dataset_parts

    with ThreadPoolExecutor(num_jobs) as ex:
        for split in dataset_parts:
            logging.info(f"Processing {split} split...")
            root = tedlium_root / split
            recordings = RecordingSet.from_dir(
                root / "sph", pattern="*.sph", num_jobs=num_jobs
            )
            stms = list((root / "stm").glob("*.stm"))
            assert len(stms) == len(recordings), (
                f"Mismatch: found {len(recordings)} "
                f"sphere files and {len(stms)} STM files. "
                f"You might be missing some parts of TEDLIUM..."
            )
            futures = []
            _parse_stm_worker = partial(_parse_stm_file, normalize_text=normalize_text)
            for stm in stms:
                futures.append(ex.submit(_parse_stm_worker, stm))

            segments = []
            for future in futures:
                segments.extend(future.result())

            supervisions = SupervisionSet.from_segments(segments)
            recordings, supervisions = fix_manifests(recordings, supervisions)

            corpus[split] = {"recordings": recordings, "supervisions": supervisions}
            validate_recordings_and_supervisions(**corpus[split])

            if output_dir is not None:
                recordings.to_file(output_dir / f"tedlium2_recordings_{split}.jsonl.gz")
                supervisions.to_file(
                    output_dir / f"tedlium2_supervisions_{split}.jsonl.gz"
                )

    return corpus


def _parse_stm_file(stm: str, normalize_text: str = "none") -> SupervisionSegment:
    """Helper function to parse a single STM file."""
    segments = []
    with stm.open() as f:
        for idx, l in enumerate(f):
            rec_id, _, _, start, end, _, *words = l.split()
            start, end = float(start), float(end)
            text = " ".join(words).replace("{NOISE}", "[NOISE]")
            if text == "ignore_time_segment_in_scoring":
                continue
            segments.append(
                SupervisionSegment(
                    id=f"{rec_id}-{idx}",
                    recording_id=rec_id,
                    start=start,
                    duration=round(end - start, ndigits=8),
                    channel=0,
                    text=normalize_text_tedlium(text, normalize_text),
                    language="English",
                    speaker=rec_id,
                )
            )
    return segments



from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

import numpy as np

from lhotse.custom import CustomFieldMixin


@dataclass
class TextExample(CustomFieldMixin):
    """
    Represents a single text example. Useful e.g. for language modeling.
    """

    text: str
    tokens: Optional[np.ndarray] = None
    custom: Optional[Dict[str, Any]] = None

    @property
    def num_tokens(self) -> Optional[int]:
        if self.tokens is None:
            return None
        return len(self.tokens)


@dataclass
class TextPairExample(CustomFieldMixin):
    """
    Represents a pair of text examples. Useful e.g. for sequence-to-sequence tasks.
    """

    source: TextExample
    target: TextExample
    custom: Optional[Dict[str, Any]] = None

    @property
    def num_tokens(self) -> Optional[int]:
        return self.source.num_tokens
